from sys import stdin
from sys import stdout

ESC = "\033["
NL = "\n"
CR = "\r"
BOLD = ESC + "1m"
CLEAR = ESC + "0m"

BLACK = ESC + "30m"
RED = ESC + "31m"
GREEN = ESC + "32m"
YELLOW = ESC + "33m"
BLUE = ESC + "34m"
MAGENTA = ESC + "35m"
CYAN = ESC + "36m"
WHITE = ESC + "37m"

BG_BLACK = ESC + "40m"
BG_RED = ESC + "41m"
BG_GREEEN = ESC + "42m"
BG_YELLOW = ESC + "43m"
BG_BLUE = ESC + "44m"
BG_MAGENTA = ESC + "45m"
BG_CYAN = ESC + "46m"
BG_WHITE = ESC + "47m"


class Console:
    @classmethod
    def cursor_up(cls):
        stdout.write(ESC + "A")

    @classmethod
    def cursor_to_left_margin(cls):
        stdout.write(CR)

    @classmethod
    def clear_line(cls):
        stdout.write(ESC + "K")

    @classmethod
    def init(cls):
        cls.clear_formating()
        stdout.write(NL + ESC + "A")

    @classmethod
    def overwrite(cls):
        cls.cursor_up()
        cls.cursor_to_left_margin()
        cls.clear_line()

    @classmethod
    def write(cls, message):
        stdout.write(message)

    @classmethod
    def writeline(cls, message):
        stdout.write(message + NL)

    @classmethod
    def read(cls):
        stdout.write(stdin.read() + NL)

    @classmethod
    def readline(cls):
        return stdin.readline()

    @classmethod
    def newline(cls):
        stdout.write(NL)

    @classmethod
    def clear_formating(cls):
        stdout.write(CLEAR)



import platform
from time import sleep
from plyer import notification
from prompt_toolkit import prompt
from prompt_toolkit.completion import Completer
from prompt_toolkit.completion import Completion
from prompt_toolkit.validation import ValidationError
from prompt_toolkit.validation import Validator
from tabulate import tabulate
from ttmr.time import Date
from ttmr.time import Alarm
from .console import BG_WHITE
from .console import BLACK
from .console import RED
from .console import Console
from .time import Timer
from .util import DefaultFormatter
from .util import add_table_border

if platform.system() == "Windows":
    import msvcrt  # pylint: disable=E0401


def display_alert(message):
    Console.write(RED)
    Console.writeline(message)
    Console.clear_formating()


def display_message(message):
    Console.clear_formating()
    Console.write(message)
    Console.newline()


def clear_line():
    Console.cursor_to_left_margin()
    Console.clear_line()


class ObjectCompleter(Completer):
    def __init__(self, objects):

        self.objects = objects

    def get_completions(self, document, complete_event):

        word_before_cursor = document.get_word_before_cursor(WORD=False)
        word_before_cursor = word_before_cursor.lower()

        for o in self.objects:
            if o.name.lower().startswith(word_before_cursor):
                yield Completion(
                    o.name, -len(word_before_cursor), display_meta=str(o.id)
                )


def format_prompt_msg(msg, min_len=10):
    _len = len(msg) if len(msg) > min_len else min_len
    INPUT_REQUEST_FMT = f"{{: >{_len}}}: "
    return INPUT_REQUEST_FMT.format(msg)


def promptor(func):
    """Prompt decorator that provides a common format
    to a prompts query.

    """

    def _promptor(input_query, *args, **kwargs):
        query = format_prompt_msg(input_query)
        return func(query, *args, **kwargs)

    return _promptor


@promptor
def text_input(input_query, default=None):
    prompt_opt = {}
    if default:
        prompt_opt["default"] = default
    return prompt(input_query, **prompt_opt)


class OptionValidator(Validator):
    def __init__(self, options):
        super(OptionValidator, self).__init__()
        self.options = options

    def validate(self, document):
        if document.text not in self.options:
            raise ValidationError(message="Must select complete row", cursor_position=0)


@promptor
def option_input(input_query, options, default=None):
    option_dict = {e.name: e for e in options}
    prompt_opt = {
        "completer": ObjectCompleter(options),
        "validator": OptionValidator(option_dict),
    }
    prompt_opt["default"] = "" if default is None else str(default)
    text = prompt(input_query, **prompt_opt)
    return option_dict.get(text, text)


class NumberValidator(Validator):
    def __init__(self, default, *args, **kwargs):
        super(NumberValidator, self).__init__(*args, **kwargs)
        self.translated_value = default

    def validate(self, document):
        text = document.text

        if text and not text.isdigit():
            i = 0

            for i, c in enumerate(text):
                if not c.isdigit():
                    break

            raise ValidationError(
                message="Duration must be a number!", cursor_position=i
            )

        self.translated_value = int(text) if text else 0


@promptor
def default_text_input(input_query, default):
    return prompt(input_query, default=default)


@promptor
def number_input(input_query, default):
    validator = NumberValidator(0)
    prompt(input_query, default=default, validator=validator)
    return validator.translated_value


@promptor
def duration_input(input_query, default=0):
    validator = NumberValidator(0)
    prompt(input_query, default=default, validator=validator)
    return validator.translated_value


class DateValidator(Validator):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.translated_value = None

    def translate(self, input_value):
        try:
            self.translated_value = Date.parse(input_value)  # , "%Y-%m-%d %H:%M:%S")
        except Exception as ex:
            raise ValidationError(message=ex)

    def validate(self, document):
        self.translate(document.text)


@promptor
def date_input(input_query, default=None):
    validator = DateValidator()
    if not default:
        display_value = Date.current().strftime("%Y-%m-%d %H:%M")
    else:
        display_value = default.strftime("%Y-%m-%d %H:%M")
    prompt(
        input_query,
        default=display_value,
        validator=validator,
        validate_while_typing=False,
    )
    return validator.translated_value


class BooleanValidator(Validator):
    def __init__(self, *args, **kwargs):
        super(BooleanValidator, self).__init__(*args, **kwargs)
        self.translated_value = False
        self.true_str = ("y", "yes", "true", "1")
        self.false_str = ("n", "no", "false", "0")

    def parse(self, input_value):
        if input_value in self.true_str:
            return True
        if input_value in self.false_str:
            return False
        raise ValidationError(message="Must enter valid reply.", cursor_position=0)

    def validate(self, document):
        self.translated_value = self.parse(document.text.lower())


@promptor
def bool_input(input_query):
    validator = BooleanValidator()
    prompt(input_query, validator=validator)
    return validator.translated_value


def write_table(data, headers, formaters=None):
    Console.write(BG_WHITE)
    Console.write(BLACK)
    Console.newline()

    fmt = DefaultFormatter()

    if formaters:
        if len(headers) != len(formaters):
            raise Exception("The number of headers and formaters do not match!")
        _data = []
        for row in data:
            _data.append(
                [
                    fmt.format(formaters[i], c) if formaters[i] else c
                    for i, c in enumerate(row)
                ]
            )

        data = _data

    table = tabulate(data, headers=headers, numalign="decimal")

    Console.write(add_table_border(table))

    Console.newline()
    Console.clear_formating()

    data.insert(0, headers)
    return data


def show_timer(start_time, alarms=None):
    """show_timer displays a timer incrementing clock in
    a HH:MM:SS format for the given start_time.

    The timer will continuously increment until the console
    registers a keyboard press.

    """
    DSPLY_FMT = " {: >10}: {}  \n\n"
    timer = Timer(start_time, alarms)

    def write_display():
        Console.write(DSPLY_FMT.format("Timer", timer.get_formatted_lapsed_time()))
        timer.check_alarms()

    write_display()

    while True:
        sleep(1)
        Console.cursor_up()
        Console.cursor_up()
        Console.cursor_to_left_margin()
        Console.clear_line()
        write_display()
        # this only works on Windows.
        if platform.system() == "Windows":
            # kbhit returns True if the keyboard is struck.
            if msvcrt.kbhit():
                break



"""
 Usage:
     ttmr current
     ttmr start
     ttmr stop
     ttmr insert
     ttmr edit --id=<INT>
     ttmr view
     ttmr summary (--day|--week) [--date=<YYYY-MM-DD>] [--csv]
     ttmr weekly [--csv]
     ttmr new  OBJTYPE
     ttmr list OBJTYPE [--inactive] [--start-date=<YYYY-MM-DD>] [--end-date=<YYYY-MM-DD>]

 Options:
   --day                      Show summary for the day.
   --week                     Show summary for the week.
   --date=<YYYY-MM-DD>        Date that defines the day/week.
   --start-date=<YYYY-MM-DD>  Date that defines the day/week.
   --end-date=<YYYY-MM-DD>    Date that defines the day/week.
   --csv                      Output csv formatted file.
   -h --help                  Show this screen.
   --version                  Show version.

"""
import codecs
import colorama
import docopt
import ttmr.config as config
from ttmr.console import GREEN
from ttmr.console import RED
from ttmr.console import Console
from ttmr.util import replace_bad_char
from . import commands

__author__ = "Mark Gemmill"
__email__ = "mark@markgemmill.com"
__version__ = "0.5.15"


colorama.init()

codecs.register_error("ttmr", replace_bad_char)


def main_wrapper(func):
    def _wrapper_():
        try:
            title = f" Task Timer v{__version__}"
            underline = "-" * (len(title) - 1)
            Console.init()
            Console.newline()
            Console.write(GREEN)
            Console.write(title)
            Console.newline()
            Console.write(f" {underline}")
            Console.clear_formating()
            Console.newline()
            Console.cursor_up()

            func()

            Console.newline()

        except KeyboardInterrupt:
            Console.clear_formating()
            Console.newline()
            Console.newline()
            Console.write(RED)
            Console.writeline("*** Task has been canceled. ***")
            Console.clear_formating()
            Console.newline()

    return _wrapper_


@main_wrapper
def main():

    args = docopt.docopt(__doc__, version="")
    conf = config.get_config(args)
    conf.cli_args = args

    Console.newline()

    if args["stop"]:
        commands.stop_time_entry(conf)
    elif args["start"]:
        commands.start_time_entry(conf)
    elif args['insert']:
        commands.new_time_entry(conf)
    elif args['edit']:
        commands.edit_time_entry(conf)
    elif args["current"]:
        commands.show_current_entry(conf)
    elif args["summary"] and args["--day"]:
        commands.day_summary(conf)
    elif args["summary"] and args["--week"]:
        commands.weekly_summary(conf)
    elif args["weekly"]:
        commands.list_current_weeks_entries(conf)
    elif args["new"] and args["OBJTYPE"] in ("entry",):
        commands.new_time_entry(conf)
    elif args["new"] and args["OBJTYPE"] in ("cat", "category"):
        commands.new_category(conf)

    elif args["list"] and args["OBJTYPE"] in ("cat", "category"):
        commands.list_categories(conf)

    elif args["new"] and args["OBJTYPE"] in (
        "prj",
        "proj",
        "inc",
        "project",
        "incident",
    ):
        commands.new_project(conf)

    elif args["list"] and args["OBJTYPE"] in (
        "prj",
        "proj",
        "inc",
        "project",
        "incident",
    ):
        commands.list_projects(conf)

    elif args["list"] and args["OBJTYPE"] in ("ent", "entry"):
        commands.list_entries(conf)

    else:
        print("Unknown command: could not complete your request.")



import os
import platform
import subprocess
from contextlib import contextmanager
from functools import wraps
from ..database import init


@contextmanager
def new_session(db_pth):
    Session = init(db_pth)
    db = Session()
    try:
        yield db
    finally:
        db.close()


def db_session(func):
    @wraps(func)
    def _db_session(cfg, *args, **kwargs):
        with new_session(cfg.db_path) as db:
            cfg.db = db
            return func(cfg, *args, **kwargs)

    return _db_session


def open_file(filepath):
    if platform.system() == "Windows":
        os.startfile(filepath)  # pylint: disable=E1101
    elif platform.system() == "Darmin":
        subprocess.Popen(["open", filepath])
    else:
        subprocess.Popen(["xdg-open", filepath])



# flake8: noqa
from .entry import show_current_entry
from .entry import start_time_entry
from .entry import stop_time_entry
from .entry import edit_time_entry
from .entry import time_entry
from .entry import new_time_entry
from .new import new_category
from .new import new_project
from .report import day_summary
from .report import list_categories
from .report import list_current_weeks_entries
from .report import list_entries
from .report import list_projects
from .report import weekly_summary
from .util import new_session



from pathlib import Path
from tempfile import NamedTemporaryFile
import csvy
from tabulate import tabulate
from .. import gui
from ..console import BG_WHITE
from ..console import BLACK
from ..console import Console
from ..database import Category
from ..database import Entry
from ..database import Project
from ..time import FMT_DATETIME
from ..util import add_table_border
from ..util import read_file
from .util import db_session
from .util import open_file


@db_session
def list_categories(conf):

    categories = Category.get_all(conf.db)

    Console.write(BG_WHITE)
    Console.write(BLACK)

    table = tabulate(
        [(c.name, c.active, c.created, c.modified) for c in categories],
        headers=("CATEGORY", "ACTIVE", "CREATED", "MODIFED"),
        numalign="decimal",
    )
    Console.write(add_table_border(table))
    Console.newline()
    Console.clear_formating()


@db_session
def list_projects(conf):
    projects = Project.get_all(conf.db)

    Console.write(BG_WHITE)
    Console.write(BLACK)

    table = tabulate(
        [(c.identifier, c.name, c.active, c.created, c.modified) for c in projects],
        headers=("IDENTIFIER", "NAME", "ACTIVE", "CREATED", "MODIFED"),
        numalign="decimal",
    )
    Console.write(add_table_border(table))
    Console.newline()
    Console.clear_formating()


def _list_entries(list_of_entries):
    entries = [
        (
            e.id,
            e.category.name,
            e.project.identifier,
            e.project.name,
            e.note,
            e.start_time,
            e.end_time,
            e.minutes,
            e.hours,
        )
        for e in list_of_entries
    ]

    total_minutes = int(sum([e[7] for e in entries]))
    total_hours = sum([e[7] for e in entries]) / 60.0

    entries.append(
        (None,"", "", "", "", entries[0][5], entries[-1][6], total_minutes, total_hours)
    )

    headers = (
        "ID",
        "CATEGORY",
        "PROJECT ID",
        "PROJECT NAME",
        "NOTE",
        "START",
        "END",
        "DURATION",
        "HOURS",
    )
    formatters = ("{:0>4}", None, None, None, None, FMT_DATETIME, FMT_DATETIME, "{:}", "{:.3}")

    gui.write_table(entries, headers, formatters)

    entries.insert(0, headers)
    return entries


@db_session
def list_entries(conf):
    _list_entries(Entry.all_between(conf.db, conf.start_date, conf.end_date))


def write_to_csv(conf, data):
    filename = Path(NamedTemporaryFile().name).name
    filepath = Path(conf.user_cache_dir, filename + ".csv")
    with csvy.writer(filepath) as csv:
        csv.writerows(data)
    open_file(filepath)


@db_session
def list_current_weeks_entries(conf):
    entries = _list_entries(Entry.current_week(conf.db))
    if conf.cli_args.get("--csv"):
        write_to_csv(conf, entries)


@db_session
def day_summary(conf):
    """
    Displays a summary of the current days total time.

    """
    rows = Entry.day_summary(conf.db)
    data = gui.write_table(
        rows,
        ("DAY", "START", "END", "MINUTES", "HOURS"),
        ("{:%Y-%m-%d}", "{:%H:%M}", "{:%H:%M}", None, "{:.3}"),
    )

    if conf.cli_args.get("--csv"):
        write_to_csv(conf, data)


@db_session
def weekly_summary(conf):
    sql = read_file("week_summary.sql")
    start = conf.today.start_of_week().to_datetime()
    end = conf.today.end_of_week().to_datetime()

    results_proxy = conf.db.execute(sql, {"start_date": start, "end_date": end})
    results = [row for row in results_proxy]

    totals = (
        "TOTALS",
        sum([r[1] for r in results]),
        sum([r[2] for r in results]),
        sum([r[3] for r in results]),
        sum([r[4] for r in results]),
        sum([r[5] for r in results]),
        sum([r[6] for r in results]),
        sum([r[7] for r in results]),
        sum([r[8] for r in results]),

    )

    results.append(totals)

    data = gui.write_table(
        results,
        ("CATEGORY", "SUN", "MON", "TUE", "WED", "THU", "FRI", "SAT", "TOTAL"),
        (None, None, None, None, None, None, None, None, None),
    )

    if conf.cli_args.get("--csv"):
        write_to_csv(conf, data)



import os
import re
import string
import sys
from importlib.resources import read_text
from . import resources


def replace_bad_char(ex):
    return (u"?", ex.start)


def get_value(input, default, type_):
    if not input:
        return type_(default)
    return type_(input)


class DefaultFormatter(string.Formatter):
    def format_field(self, value, format_spec):
        try:
            return super().format_field(value, format_spec)
        except:
            return "" if value is None else str(value)


def find_data_file(filename):
    if getattr(sys, "frozen", False):
        # The application is frozen
        datadir = os.path.dirname(sys.executable)
    else:
        # The application is not frozen
        # Change this bit to match where you store your data files:
        datadir = os.path.dirname(__file__)
    return os.path.join(datadir, "sql", filename)


def read_file(filename):
    return read_text(resources, filename)


def add_table_border(text, border=1):
    lines = [(" " * border) + line for line in text.split("\n")]
    length = max([len(a) for a in lines])

    return "\n".join([l.ljust(length) for l in lines])


def regex_factory(regex):
    """Convenience wrapper around regex with named capture groups.
    """
    reg = re.compile(regex, re.I)

    def _match(text):
        m = reg.match(text)
        if m:
            return m.groupdict()
        return None

    return _match



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ttmr', 'ttmr.commands', 'ttmr.resources']

package_data = \
{'': ['*'], 'ttmr': ['sql/*']}

install_requires = \
['appdirs>=1.4.3,<2.0.0',
 'arrow>=0.15.5,<0.16.0',
 'colorama>=0.4.1,<0.5.0',
 'csvy>=0.1.0-alpha.0,<0.2.0',
 'docopt-ng>=0.7.2,<0.8.0',
 'pendulum>=2.0.5,<3.0.0',
 'plyer>=1.4.2,<2.0.0',
 'prompt-toolkit>=3.0.3,<4.0.0',
 'pytest-cov>=2.8.1,<3.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'sqlalchemy>=1.3.13,<2.0.0',
 'tabulate>=0.8.6,<0.9.0',
 'tomlkit>=0.5.8,<0.6.0',
 'typer>=0.1.0,<0.2.0']

extras_require = \
{':sys_platform == "linux"': ['dbus-python>=1.2.16,<2.0.0']}

entry_points = \
{'console_scripts': ['ttmr = ttmr.cli:main', 'ttmrc = ttmr.clic:main']}

setup_kwargs = {
    'name': 'ttmr',
    'version': '0.5.15',
    'description': '',
    'long_description': None,
    'author': 'Mark Gemmill',
    'author_email': 'gitlab@markgemmill.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import typer


app = typer.Typer()


@app.command()
def current():
    pass

@app.command()
def start():
    pass


@app.command()
def stop():
    pass


@app.command()
def view():
    pass


@app.command()
def summary(day: bool = False, week: bool = False):
    print("calling summary...")
    print(f"day: {day}")
    print(f"week: {week}")


@app.command()
def weekly():
    pass


@app.command()
def new():
    pass


@app.command()
def list():
    pass


def main():
    app()

